from __future__ import annotations

import functools
import inspect
from collections.abc import Mapping, MutableMapping
from typing import Any, Iterator, Optional, Union

from octoflow.utils.objects import create_object

__all__ = [
    "Config",
]


class ConfigWrapper:
    def __init__(
        self,
        wrapped,
        config: Config,
        name: Optional[str] = None,
    ):
        if name is None:
            name = getattr(wrapped, "__name__", None)
        if name is None:
            name = getattr(wrapped, "__qualname__", None)
        if name is None:
            msg = "builder must have a name"
            raise ValueError(msg)
        self.wrapped = wrapped
        self.signature = inspect.signature(wrapped)
        self.name = name
        self.config = config

    def _update_params_from_config(self, args):
        if self.name not in self.config:
            # function or class name or alias is not in config
            return args
        config = self.config[self.name]
        # update args
        for k in config:
            if k in args:  # and args[k] is not __placeholder__
                # already set - do not override
                continue
            args[k] = config[k]
        return args

    def get_params(self, *args, **kwargs):
        args = self.signature.bind_partial(*args, **kwargs)
        args = args.arguments
        args = self._update_params_from_config(args)
        args = self.signature.bind_partial(**args)
        args.apply_defaults()
        args = args.arguments
        return args

    def __call__(self, *args, **kwargs):
        return self.wrapped(**self.get_params(*args, **kwargs))


class FrozenConfig(Mapping):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._data = dict(*args, **kwargs)

    def __getitem__(self, key: str) -> Any:
        value = self._data[key]
        if isinstance(value, dict):
            if "$type" in value:
                # value is an object
                params = dict(value.items())
                cls = params.pop("$type")
                partial = params.pop("$partial", False)
                kwargs = FrozenConfig(params).to_dict()
                return create_object(
                    cls,
                    partial,
                    **kwargs,
                )
            else:
                return FrozenConfig(value)
        return value

    def __getattr__(self, name: str) -> Any:
        if name in self:
            return self[name]
        return None

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator:
        yield from self._data

    def to_dict(self) -> dict:
        out = {}
        for k, v in self.items():
            out[k] = v.to_dict() if isinstance(v, FrozenConfig) else v
        return out

    def __reduce__(self) -> tuple:
        return (self.__class__, (self._data,))


class Config(FrozenConfig, MutableMapping):
    def __setitem__(self, key: str, value: Any) -> None:
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def wraps(
        self,
        wrapped: Union[type, callable, None] = None,
        **kwargs,
    ) -> Union[ConfigWrapper, functools.partial]:
        if isinstance(wrapped, str):
            return functools.partial(
                self.wraps,
                name=wrapped,
            )
        elif wrapped is None:
            # empty builder call
            return functools.partial(
                self.wraps,
                **kwargs,
            )
        wrapped: ConfigWrapper = ConfigWrapper(
            wrapped,
            self,
            name=kwargs.get("name", None),
        )
        if isinstance(wrapped, type):  # if class
            orig_init = wrapped.__init__

            @functools.wraps(wrapped.__init__)
            def __init__(self, *args, **kwargs):  # noqa: N807
                orig_init(self, **wrapped.get_params(*args, **kwargs))

            wrapped.__init__ = __init__
            return wrapped

        @functools.wraps(wrapped)
        def builder_wrapper(*args, **kwargs):
            return wrapped(*args, **kwargs)

        return builder_wrapper



from __future__ import annotations

from typing import List, Optional

from octoflow.tracking.base import SessionMixin, create_engine
from octoflow.tracking.experiment import Experiment
from octoflow.tracking.run import Run


class Client(SessionMixin):
    def __init__(self, tracking_uri: Optional[str] = None) -> None:
        """
        Client for interacting with the database.

        Parameters
        ----------
        tracking_uri : str, optional
            Database URI, by default None.
        """
        # if tracking_uri is None -> defaults to 'sqlite:///:memory:'
        self.tracking_uri = tracking_uri
        self.engine = create_engine(tracking_uri)

    def create_experiment(
        self,
        name: str,
        description: Optional[str] = None,
        artifact_uri: Optional[str] = None,
        return_if_exist: bool = True,
    ) -> Experiment:
        """
        Create an experiment.

        Parameters
        ----------
        name : str
            Experiment name.
        description : str, optional
            Experiment description, by default None.
        return_if_exist : bool, optional
            Return the experiment if it already exists, by default True.

        Returns
        -------
        Experiment
            Experiment object.
        """
        with self.session():
            expr = self.get_experiment_by_name(name)
            if expr is not None:
                if return_if_exist:
                    return expr
                msg = f"experiment with name '{name}' already exists"
                raise ValueError(msg)
            expr = Experiment(
                name=name,
                description=description,
                artifact_uri=artifact_uri,
            )
        return expr

    def get_experiment_by_name(self, name: str) -> Experiment:
        """
        Get an experiment by name.

        Parameters
        ----------
        name : str
            Experiment name.

        Returns
        -------
        Experiment
            Experiment object.
        """
        with self.session() as session:
            expr = session.query(Experiment).filter_by(name=name).first()
        return expr

    def list_experiments(self) -> List[Experiment]:
        """
        List all experiments.

        Returns
        -------
        List[Experiment]
            List of experiments.
        """
        with self.session() as session:
            exprs = session.query(Experiment).all()
        return exprs

    def get_experiment(self, id: int) -> Experiment:
        """
        Get an experiment by id.

        Parameters
        ----------
        id : int
            Experiment id.

        Returns
        -------
        Experiment
            Experiment object.
        """
        if id is None:
            msg = "experiment id cannot be None"
            raise ValueError(msg)
        with self.session() as session:
            try:
                expr = session.query(Experiment).get(id)
            except ValueError:
                msg = f"experiment with id '{id}' does not exist"
                raise ValueError(msg) from None
        return expr

    def get_run(self, id: int) -> Run:
        """
        Get a run by id.

        Parameters
        ----------
        id : int
            Run id.

        Returns
        -------
        Run
            Run object.
        """
        if id is None:
            msg = "run id cannot be None"
            raise ValueError(msg)
        with self.session() as session:
            try:
                run = session.query(Run).get(id)
            except ValueError:
                msg = f"run with id '{id}' does not exist"
                raise ValueError(msg) from None
        return run



from typing import Generic, TypeVar

ArrowType = TypeVar("ArrowType")


class PyArrowWrapper(Generic[ArrowType]):
    def __init__(self, wrapped: ArrowType) -> None:
        self._wrapped = wrapped

    def to_pyarrow(self) -> ArrowType:
        return self._wrapped



import importlib


class Package:
    def __init__(self, name: str, modules: list):
        self.name = name
        self.modules = modules

    def import_modules(self):
        for module in self.modules:
            if isinstance(module, str):
                module = {"name": module}
            if not isinstance(module, dict):
                msg = f"expected str or dict, got {type(module).__name__}"
                raise TypeError(msg)
            name, package = module["name"], module.get("package")
            if package is None:
                package = __package__
            try:
                module = importlib.import_module(
                    name=name,
                    package=package,
                )
            except Exception:
                msg = f"failed to import '{name}' from '{package}' in '{self.name}'"
                raise ImportError(msg) from None
            if not hasattr(module, "__all__"):
                continue
            for name in module.__all__:
                subpackage = getattr(module, name)
                if not isinstance(subpackage, Package):
                    continue
                subpackage.import_modules()



"""Run model.

This module contains the run model.
"""

from __future__ import annotations

import uuid
from datetime import datetime
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

from packaging.version import Version
from sqlalchemy import (
    DateTime,
    ForeignKey,
    Integer,
    String,
    Text,
)
from sqlalchemy.orm import Mapped, aliased, mapped_column
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound

from octoflow.tracking import artifact, value_utils
from octoflow.tracking.artifact import Artifact
from octoflow.tracking.base import Base
from octoflow.tracking.value import Value, ValueType
from octoflow.tracking.variable import Variable, VariableType

__all__ = [
    "Run",
]

NOT_SPECIFIED = object()


def generate_uuid() -> str:
    return str(uuid.uuid4())


class Run(Base):
    """Run model."""

    __tablename__ = "run"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    experiment_id: Mapped[int] = mapped_column(Integer, ForeignKey("experiment.id", ondelete="CASCADE"))
    name: Mapped[str] = mapped_column(String, nullable=False)
    description: Mapped[Optional[str]] = mapped_column(Text, nullable=True)
    created_at: Mapped[datetime] = mapped_column(DateTime, nullable=False, default=datetime.utcnow)
    uuid: Mapped[str] = mapped_column(String(36), nullable=False, unique=True, default=generate_uuid)

    def _log_value(
        self,
        key: str,
        value: ValueType,
        *,
        step: Optional[Value] = None,
        type: Optional[VariableType] = None,
    ) -> Value:
        """Log a value."""
        parent_id = None if step is None else step.variable_id
        step_id = None if step is None else step.id
        with self.session():
            # get or create variable
            variable = Variable.get(
                key=key,
                experiment_id=self.experiment_id,
                parent_id=parent_id,
                type=type,
                create=True,  # create variable if it does not exist
            )
            value = Value(
                run_id=self.id,
                variable_id=variable.id,
                value=value,
                step_id=step_id,
            )
        return value

    def _log_values(
        self,
        *,
        values: Dict[str, ValueType],
        step: Optional[Value] = None,
        type: Optional[VariableType] = None,
        prefix: Optional[str] = None,
    ) -> List[Value]:
        """Log multiple values."""
        result = []
        prefix = "" if prefix is None else str(prefix).strip()
        for key, value in value_utils.flatten(values).items():
            if len(prefix) > 0:
                d = "." if prefix[-1] != "." else ""
                key = f"{prefix}{d}{key}"
            value = self._log_value(
                key=key,
                value=value,
                step=step,
                type=type,
            )
            result.append(value)
        return result

    def log_param(
        self,
        key: str,
        value: ValueType,
        step: Optional[Value] = None,
    ) -> Value:
        """Log a parameter.

        Parameters
        ----------
        key : str
            Parameter key.
        value : ValueType
            Parameter value.
        step : Optional[Value], optional
            Step value, by default None.

        Returns
        -------
        Value
            Value instance.
        """
        return self._log_value(
            key=key,
            value=value,
            step=step,
            type=VariableType.parameter,
        )

    def log_params(
        self,
        values: Dict[str, ValueType],
        step: Optional[Value] = None,
        prefix: Optional[str] = None,
    ) -> List[Value]:
        """Log multiple parameters.

        Parameters
        ----------
        values : Dict[str, ValueType]
            Dictionary of values.
        step : Optional[Value], optional
            Step value, by default None.

        Returns
        -------
        List[Value]
            List of values.
        """
        return self._log_values(
            values=values,
            step=step,
            type=VariableType.parameter,
            prefix=prefix,
        )

    def get_param(
        self,
        key: str,
        default: Optional[ValueType] = NOT_SPECIFIED,
        *,
        step: Optional[Value] = NOT_SPECIFIED,
    ) -> ValueType:
        """Get a parameter.

        Parameters
        ----------
        key : str
            Parameter key.
        default : Optional[ValueType], optional
            Default value, by default NOT_SPECIFIED.
        step : Optional[Value], optional
            Step value, by default NOT_SPECIFIED.

        Returns
        -------
        Value
            Value instance.
        """
        with self.session() as session:
            q = (
                session.query(Value)
                .join(Variable)
                .filter(
                    Value.run_id == self.id,
                    Variable.key == key,
                    Variable.type == VariableType.parameter,
                )
            )
            if step is None:
                q = q.filter(Value.step_id.is_(None))
            if step is not NOT_SPECIFIED:
                q = q.filter(Value.step_id == step.id)
            try:
                value = q.one().value
            except NoResultFound as e:
                if default is NOT_SPECIFIED:
                    msg = f"parameter with key '{key}' does not exist"
                    raise ValueError(msg) from e
                value = default
            except MultipleResultsFound as e:
                msg = f"multiple parameters with key '{key}' exist"
                raise ValueError(msg) from e
        return value

    def log_metric(
        self,
        key: str,
        value: ValueType,
        step: Optional[Value] = None,
    ) -> Value:
        """Log a metric.

        Parameters
        ----------
        key : str
            Metric key.
        value : ValueType
            Metric value.
        step : Optional[Value], optional
            Step value, by default None.

        Returns
        -------
        Value
            Value instance.
        """
        return self._log_value(
            key=key,
            value=value,
            step=step,
            type=VariableType.metric,
        )

    def log_metrics(
        self,
        values: Dict[str, ValueType],
        step: Optional[Value] = None,
        prefix: Optional[str] = None,
    ) -> List[Value]:
        """Log multiple metrics.

        Parameters
        ----------
        values : Dict[str, ValueType]
            Dictionary of values.
        """
        return self._log_values(
            values=values,
            step=step,
            type=VariableType.metric,
            prefix=prefix,
        )

    def log_artifact(
        self,
        key: str,
        value: Any = None,
        *,
        version: Union[str, Version, None] = None,
        handler_type: Optional[str] = None,
        save: bool = True,
        **kwargs: Dict[str, Any],
    ) -> Artifact:
        """Log an artifact.

        Parameters
        ----------
        key : str
            Artifact key.
        value : Optional[Any], optional
            Artifact value, by default None.
        version : Union[str, Version, None], optional
            Artifact version, by default None.
        path : Union[str, Path]
            Artifact path.
        handler_type : Union[str, None]
            Artifact handler (name), by default None.
        save : bool, optional
            Whether to save the artifact, by default True.
        kwargs : Dict[str, Any]
            Additional keyword arguments.

        Returns
        -------
        Value
            Value instance.
        """
        if handler_type is None and value is not None:
            handler_type = artifact.get_handler_type_by_object(value)
        elif isinstance(handler_type, str):
            # validate if a handler for type exists
            handler_type = artifact.get_handler_type(handler_type)
        with self.session():
            a = Artifact(
                run_id=self.id,
                key=key,
                handler_type=handler_type,
                version=version,
            )
        if save:
            a.save(value, **kwargs)
        return a

    def get_artifact(
        self,
        key: str,
        version: Union[str, Version, None] = None,
    ) -> Artifact:
        """Load an artifact.

        Parameters
        ----------
        key : str
            Artifact key.
        version : Union[str, Version, None], optional
            Artifact version, by default None.

        Returns
        -------
        Any
            Artifact value.
        """
        with self.session() as session:
            q = session.query(Artifact).filter(
                Artifact.run_id == self.id,
                Artifact.key == key,
            )
            if version is not None and str(version) == "latest":
                a = q.order_by(Artifact.version.desc()).first()
            else:
                if version is not None:
                    q = q.filter(Artifact.version == str(version))
                try:
                    a = q.one_or_none()
                except MultipleResultsFound as e:
                    if version is None:
                        # ask user to specify version to narrow down the search
                        msg = f"multiple artifacts with key '{key}' exist, please specify version"
                        raise ValueError(msg) from e
                    raise e
        if a is None:
            msg = f"artifact with key '{key}' does not exist"
            raise ValueError(msg)
        return a

    def get_raw_logs(self) -> List[Tuple[int, int, int, str, Any]]:
        """Get all logs of the run.

        Returns
        -------
        List[Tuple[int, int, int, bool, str, Any]]
            List of logs.
        """
        with self.session() as session:
            child_value = aliased(Value)
            # get all values of this run and their corresponding variable keys
            values = (
                session.query(
                    Value.run_id,
                    Value.id,
                    Value.step_id,
                    child_value.id.is_not(None).label("is_step"),
                    Variable.key,
                    Value.value,
                )
                .select_from(Value)
                .join(Variable, Variable.id == Value.variable_id)
                .outerjoin(child_value, child_value.step_id == Value.id)
                .filter(Value.run_id == self.id)
                .distinct(Value.id)
                .all()
            )
        return values

    def get_logs(self) -> value_utils.ValueTree:
        """Get all logs of the run.

        Returns
        -------
        ValueTree
            List of logs.
        """
        values = self.get_raw_logs()
        trees = value_utils.build_trees(values)
        return trees[self.id]

    def match(self, partial: bool = True) -> Generator[int, None, None]:
        """Checks whether the run with same values for parameter typed variables exists in the database.

        Parameters
        ----------
        partial : bool, optional
            Whether to check for partial match, by default True.

        Yields
        ------
        Generator[int, None, None]
            Generator of run IDs.
        """
        with self.session() as session:
            # get all values of this run and their corresponding variable keys
            child_value = aliased(Value)
            values = (
                session.query(
                    Value.run_id,
                    Value.id,
                    Value.step_id,
                    child_value.id.is_not(None).label("is_step"),
                    Variable.key,
                    Value.value,
                )
                .select_from(Value)
                .join(Variable, Variable.id == Value.variable_id)
                .outerjoin(child_value, child_value.step_id == Value.id)
                .filter(
                    Variable.experiment_id == self.experiment_id,
                    Variable.type == VariableType.parameter,
                )
                .distinct(Value.id)
                .all()
            )
            # build up a tree of values
            trees = value_utils.build_trees(values)
        # check whether there is a run with same values
        if self.id not in trees:
            msg = f"run with id '{self.id}' does not exist"
            raise ValueError(msg)
        this = trees.pop(self.id).normalize()
        for run_id, other in trees.items():
            other = other.normalize()
            if value_utils.equals(this, other, partial=partial):
                yield run_id



from __future__ import annotations

import functools
import weakref
from collections import defaultdict
from typing import Dict, Generator, Iterable, MutableMapping, MutableSequence, Optional, Set, TypeVar
from typing import MutableSet as MutableSetType

K, V = TypeVar("K"), TypeVar("V")


class EventTarget:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._event_listeners: Dict[str, Set[callable]] = defaultdict(set)

    def add_event_listener(self, type: str, listener: callable):
        self._event_listeners[type].add(listener)

    def remove_event_listener(self, type: str, listener: callable):
        self._event_listeners[type].remove(listener)

    def dispatch_event(self, event: str):
        for callback in self._event_listeners[event]:
            callback()


class MutableCollection(EventTarget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._parent_ref: MutableCollection = None

    @property
    def _parent(self) -> Optional[MutableCollection]:
        if self._parent_ref is None:
            return None
        return self._parent_ref()

    def set_parent(self, parent: MutableCollection) -> MutableCollection:
        self._parent_ref = weakref.ref(parent)
        return self

    def changed(self):
        if self._parent is not None:
            self._parent.changed()
        self.dispatch_event("change")

    def coerce(self, value: any):
        if isinstance(value, (list, MutableList)):
            return MutableList(value).set_parent(self)
        elif isinstance(value, (dict, MutableDict)):
            return MutableDict(value).set_parent(self)
        elif isinstance(value, (set, MutableSet)):
            return MutableSet(value).set_parent(self)
        return value


class MutableDict(MutableCollection, MutableMapping[K, V]):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._data = {}
        self.update(dict(*args, **kwargs))

    def __getitem__(self, key: K) -> V:
        return self._data[key]

    def __setitem__(self, key: K, value: V):
        self._data[key] = self.coerce(value)
        self.changed()

    def __delitem__(self, key: K):
        del self._data[key]
        self.changed()

    def __iter__(self) -> Generator[K, None, None]:
        yield from self._data

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return repr(self._data)


class MutableList(MutableCollection, MutableSequence[V]):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._data = []
        self.extend(list(*args, **kwargs))

    def __getitem__(self, key: int) -> V:
        return self._data[key]

    def __setitem__(self, key: int, value: V):
        self._data[key] = self.coerce(value)
        self.changed()

    def insert(self, index: int, value: V):
        self._data.insert(index, self.coerce(value))
        self.changed()

    def __delitem__(self, key: int):
        del self._data[key]
        self.changed()

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return repr(self._data)

    @functools.wraps(list.sort)
    def sort(self, *args, **kwargs) -> None:
        self._data.sort(*args, **kwargs)
        self.changed()


class MutableSet(MutableCollection, MutableSetType[V]):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._data = set()
        self.update(set(*args, **kwargs))

    def add(self, item: V):
        self._data.add(self.coerce(item))
        self.changed()

    def update(self, *s: Iterable[V]) -> None:
        self._data.update({self.coerce(i) for i in set(*s)})
        self.changed()

    def discard(self, item: V):
        self._data.discard(item)
        self.changed()

    def __contains__(self, item: V) -> bool:
        return item in self._data

    def __iter__(self) -> Generator[V, None, None]:
        yield from self._data

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return repr(self._data)



from octoflow.tracking.artifact import Artifact
from octoflow.tracking.base import Base
from octoflow.tracking.client import Client
from octoflow.tracking.experiment import Experiment
from octoflow.tracking.run import Run
from octoflow.tracking.value import Value
from octoflow.tracking.variable import Variable, VariableType

__all__ = [
    "Base",
    "Client",
    "Experiment",
    "Run",
    "Value",
    "Variable",
    "VariableType",
    "Artifact",
]



from typing import Optional

from octoflow import logging
from octoflow.config import Config
from octoflow.core.plugins import Package
from octoflow.tracking import Client, Experiment, Run, Value

__version__ = "0.0.13"

__all__ = [
    "Client",
    "Experiment",
    "Run",
    "Value",
    "Config",
    "LoggingFactory",
    "logging",
]

config = Config({
    "resources": {
        "path": "~/.octoflow",
    },
    "logging": {
        "level": "INFO",
        "format": "%(asctime)s %(levelname)s %(name)s %(message)s",
    },
})


logger = logging.get_logger(
    name=next(iter(__package__.split("."))),
    level=config.logging.level,
    formatter=config.logging.format,
)


default_plugins_package: Optional[Package]

try:
    from octoflow_plugins import package as default_plugins_package  # type: ignore
except ImportError:
    default_plugins_package = None

if default_plugins_package is not None:
    default_plugins_package.import_modules()



from __future__ import annotations

from datetime import datetime
from distutils.version import Version
from pathlib import Path
from typing import Any, Type, Union

from sqlalchemy import (
    DateTime,
    ForeignKey,
    Index,
    Integer,
    String,
    text,
    types,
)
from sqlalchemy.orm import Mapped, mapped_column

from octoflow.tracking.artifact import handler
from octoflow.tracking.artifact.handler import (
    ArtifactHandler,
    get_handler_type,
    get_handler_type_by_object,
    list_handler_types,
)
from octoflow.tracking.base import Base

__all__ = [
    "Artifact",
    "handler",
    "get_handler_type",
    "get_handler_type_by_object",
    "list_handler_types",
]


class ArtifactHandlerType(types.TypeDecorator):
    """A SQLAlchemy type decorator for ArtifactHandler."""

    impl = types.Unicode

    cache_ok = True

    def process_bind_param(self, value: Type[ArtifactHandler], dialect) -> str:  # noqa: PLR6301
        """Gets the name from the handler type."""
        if not issubclass(value, ArtifactHandler):
            msg = f"'{value}' is not a subclass of 'ArtifactHandler'"
            raise ValueError(msg)
        return value.name

    def process_result_value(self, value: str, dialect) -> Type[ArtifactHandler]:  # noqa: PLR6301
        """Gets the handler type for the given name."""
        return get_handler_type(value)


class VersionType(types.TypeDecorator):
    """A SQLAlchemy type decorator for Version."""

    impl = types.Unicode

    cache_ok = True

    def process_bind_param(self, value: Union[str, Version], dialect) -> str:  # noqa: PLR6301
        """Converts the value to a string."""
        if not isinstance(value, Version):
            value = Version(value)
        return str(value)

    def process_result_value(self, value: str, dialect) -> Version:  # noqa: PLR6301
        """Converts the value to a Version."""
        return Version(value)


class Artifact(Base):
    __tablename__ = "artifact"
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    run_id: Mapped[int] = mapped_column(Integer, ForeignKey("run.id"), nullable=False)
    key: Mapped[str] = mapped_column(String(255), nullable=False)
    version: Mapped[Version] = mapped_column(String(128), nullable=False, default="0.0.0")
    handler_type: Mapped[Type[ArtifactHandler]] = mapped_column(ArtifactHandlerType, nullable=False)
    timestamp: Mapped[datetime] = mapped_column(DateTime, nullable=False, default=datetime.utcnow)

    __table_args__ = (
        Index(
            "ix_artifact_run_id_key",
            "run_id",
            "key",
            "version",
            unique=True,
        ),
    )

    @property
    def path(self) -> Path:
        if hasattr(self, "_path") and self._path is not None:
            return self._path
        with self.session() as session:
            query = text("SELECT r.experiment_id, r.uuid FROM run as r WHERE r.id = :run_id")
            exper_id, run_uuid = session.execute(query, {"run_id": self.run_id}).one()
            query = text("SELECT e.artifact_uri FROM experiment as e WHERE e.id = :experiment_id")
            arifact_uri = session.execute(query, {"experiment_id": exper_id}).scalar()
            path: Path = Path(arifact_uri) / run_uuid / f"{self.key}-v{self.version}"
            if not path.exists():
                path.mkdir(parents=True)
            self._path = path
        return self._path

    @property
    def handler(self) -> ArtifactHandler:
        """Return the handler for this artifact."""
        if not hasattr(self, "_handler") or self._handler is None:
            self._handler = self.handler_type(
                self.path,
            )
        return self._handler

    def load(self) -> Any:
        """
        Load the artifact.

        Returns
        -------
        Any
            The loaded artifact.
        """
        return self.handler.load()

    def save(self, obj: Any, **kwargs) -> None:
        """
        Save an object to the artifact.

        Parameters
        ----------
        obj : Any
            The object to save.
        kwargs : dict
            Additional keyword arguments.
        """
        self.handler.save(obj, **kwargs)

    def exists(self) -> bool:
        """
        Return True if the artifact exists.

        Returns
        -------
        bool
            True if the artifact exists.
        """
        return self.handler.exists()

    def unlink(self) -> None:
        """
        Unlink/delete the artifact.

        Notes
        -----
        This will only delete the artifact, not the database record.

        Returns
        -------
        None
            None
        """
        self.handler.unlink()

    def delete(self, unlink: bool = False) -> None:
        """
        Delete the artifact.

        Parameters
        ----------
        unlink : bool, optional
            If True, unlink the artifact, by default False.

        Notes
        -----
        This will delete the database record and unlink the artifact if `unlink` is True.

        Returns
        -------
        None
            None
        """
        try:
            # delete the database record
            super().delete()
            if unlink:
                # unlink the artifact
                self.handler.unlink()
        except ValueError as ex:
            raise ex



from __future__ import annotations

from typing import Any, Tuple, Union

import pyarrow.dataset as ds
from pyarrow.dataset import field as pyarrow_field
from pyarrow.dataset import scalar as pyarrow_scalar

from octoflow.data.base import PyArrowWrapper

__all__ = [
    "Expression",
]

BaseExpression = PyArrowWrapper[ds.Expression]


class Expression(BaseExpression):
    def __init__(self, expression: Union[Expression, ds.Expression]):
        if isinstance(expression, Expression):
            expression = expression._wrapped
        super().__init__(expression)

    def __eq__(self, other: Any) -> Expression:
        return Expression(self._wrapped == Expression(other)._wrapped)

    def __ne__(self, other: Any) -> Expression:
        return Expression(self._wrapped != Expression(other)._wrapped)

    def __lt__(self, other: Any) -> Expression:
        return Expression(self._wrapped < Expression(other)._wrapped)

    def __le__(self, other: Any) -> Expression:
        return Expression(self._wrapped <= Expression(other)._wrapped)

    def __gt__(self, other: Any) -> Expression:
        return Expression(self._wrapped > Expression(other)._wrapped)

    def __ge__(self, other: Any) -> Expression:
        return Expression(self._wrapped >= Expression(other)._wrapped)

    def __and__(self, other: Any) -> Expression:
        return Expression(self._wrapped & Expression(other)._wrapped)

    def __or__(self, other: Any) -> Expression:
        return Expression(self._wrapped | Expression(other)._wrapped)

    def __invert__(self) -> Expression:
        return Expression(~self._wrapped)

    def is_nan(self) -> Expression:
        return Expression(self._wrapped.is_nan())

    def is_null(self, nan_is_null: bool = False):
        return Expression(self._wrapped.is_null(nan_is_null=nan_is_null))

    def is_valid(self) -> Expression:
        return Expression(self._wrapped.is_valid())

    def isin(self, other: Expression) -> Expression:
        return Expression(self._wrapped.isin(Expression(other)._wrapped))

    def equals(self, other: Expression) -> Expression:
        return Expression(self._wrapped.equals(Expression(other)._wrapped))

    @classmethod
    def field(cls, *name_or_index: Tuple[str]) -> Expression:
        field = pyarrow_field(*name_or_index)
        return cls(field)

    @classmethod
    def scalar(cls, value: Any) -> Expression:
        scaler = pyarrow_scalar(value)
        return cls(scaler)

    def __hash__(self) -> int:
        return hash(self._wrapped)

    def __repr__(self) -> str:
        return f"Expression({self._wrapped})"



from __future__ import annotations

import copy
from typing import Any, Dict, List, Mapping, Optional


class ValueTuple(tuple):
    """A value tuple."""

    def __new__(cls, key: str, value: Any, is_step: Optional[bool] = None) -> ValueTuple:
        return super(ValueTuple, cls).__new__(cls, (key, value))

    def __init__(self, key: str, value: Any, is_step: Optional[bool]):
        self.is_step = is_step

    @property
    def key(self) -> str:
        """Get the key."""
        return self[0]

    @property
    def value(self) -> Any:
        """Get the value."""
        return self[1]

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ValueTuple):
            if isinstance(other, tuple):
                other = ValueTuple(*other)
            else:
                return False
        return self.key == other.key and self.value == other.value

    def __hash__(self) -> int:
        return hash((self.key, self.value))

    def __repr__(self) -> str:
        return f"({self.key}, {self.value})"


class ValueTree(Mapping[int, "ValueTree"]):
    """A value tree."""

    def __init__(
        self,
        data: Mapping[int, dict],
        id2value: Dict[int, ValueTuple],
        root: bool = True,
    ):
        super(ValueTree, self).__init__()
        self.data = data
        self.id2value = id2value
        self.root = root

    def __getitem__(self, key: int) -> ValueTree:
        subtree = self.data[key]
        return ValueTree(subtree, id2value=self.id2value, root=False)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def select(
        self,
        *keys: str,
    ) -> List[dict]:
        """
        Select values from the tree.

        Parameters
        ----------
        keys : str
            The keys to select.

        Returns
        -------
        List[dict]
            A list of dictionaries containing the selected values.
        """
        if len(keys) == 0 or len(self) == 0:
            # either the keys are empty or the tree is empty
            return []
        record = {}
        remaining_keys = set(keys)
        for value_id, subtree in self.items():
            # add all the values that are not steps if they are in the keys
            value = self.id2value[value_id]
            if value.is_step:
                continue
            if len(subtree) != 0:
                # child is a parent
                msg = f"variable '{value.key}' is not marked with 'is_step'"
                raise ValueError(msg)
            if value.key in keys:
                # parent is the key
                remaining_keys = remaining_keys - {value.key}
                record[value.key] = value.value
        if len(remaining_keys) == 0:
            # no keys left
            return [record]
        output = []
        for value_id, subtree in self.items():
            value = self.id2value[value_id]
            if not value.is_step:
                continue
            out = copy.deepcopy(record)
            has_key = value.key in remaining_keys
            if has_key:
                out[value.key] = value.value
            subtree_remaining_keys = remaining_keys - {value.key}
            if len(subtree_remaining_keys) == 0:
                # remaining key was found as step variable
                output.append(out)
                continue
            subtree_selection = subtree.select(*subtree_remaining_keys)
            if len(subtree_selection) > 0 and not has_key:
                # looks like user selected a key without their parent step value
                msg = f"missing '{value.key}'"
                raise ValueError(msg)
            for subtree_out in subtree_selection:
                # there are keys left but the subtree does not contain them
                subtree_out.update(out)
                output.append(subtree_out)
        return output

    def normalize(self) -> dict:
        """
        Normalize a value tree by removing IDs and replacing them with key-value pairs.

        Returns
        -------
        dict
            The normalized value tree.
        """
        normalized_tree = {}
        for parent_id, subtree in self.items():
            if parent_id is None:
                msg = "key in ValueTree must not be None"
                raise ValueError(msg)
            parent = self.id2value[parent_id]
            if not isinstance(subtree, ValueTree):
                msg = f"subtree must be of type ValueTree, got {type(subtree).__name__}"
                raise ValueError(msg)
            normalized_tree[parent] = subtree.normalize()
        return normalized_tree

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data})"


def build_trees(values: List) -> Dict[int, ValueTree]:
    """
    Build a value tree from a list of values.

    Parameters
    ----------
    values : List
        A list of values containing tuples of (run_id, value_id, step_id, is_step, key, value).

    Returns
    -------
    Dict[int, dict]
        A dictionary representing the value tree, where the keys are run IDs and the values are nested dictionaries.
    """
    trees: dict = {}
    id2value = {}
    for run_id, value_id, step_id, is_step, key, value in values:
        id2value[value_id] = ValueTuple(key, value, is_step=is_step)
        if run_id not in trees:
            trees[run_id] = {}
        root: dict = trees[run_id]
        if step_id not in root or root[step_id] is None:
            root[step_id] = {}
        node = root[value_id] if value_id in root else root.setdefault(value_id, {})
        root[step_id][value_id] = node
    return {
        run_id: ValueTree(
            root[None],
            id2value=id2value,
        )
        for run_id, root in trees.items()
    }


def equals(
    p: dict,
    q: dict,
    /,
    partial: bool = True,
) -> bool:
    """
    Check whether two dictionaries are equal.

    Parameters
    ----------
    p : dict
        The first dictionary to compare.
    q : dict
        The second dictionary to compare.
    partial : bool, optional
        If True, performs a partial comparison where missing keys in one dictionary are ignored (default is True).

    Returns
    -------
    bool
        True if the dictionaries are equal, False otherwise.

    Raises
    ------
    ValueError
        If either `p` or `q` is not a dictionary.
    """
    if not partial:
        if p.keys() != q.keys():
            return False
        return all(equals(value, q[key]) for key, value in p.items())
    if not isinstance(p, dict) or not isinstance(q, dict):
        msg = f"cannot compare {type(p).__name__} and {type(q).__name__} types"
        raise ValueError(msg)
    for key, val_a in p.items():
        if key not in q:
            return False
        val_b = q[key]
        if not equals(val_a, val_b):
            return False
    return True


def flatten(
    data: Dict[str, Any],
    parent_key: str = "",
    separator: str = ".",
) -> dict[str, Any]:
    """
    Flatten a nested dictionary.

    Parameters
    ----------
    data : Dict[str, Any]
        The nested dictionary to flatten.

    Returns
    -------
    dict[str, Any]
        The flattened dictionary.
    """
    items = []
    for key, value in data.items():
        # escape dots
        new_key = parent_key + separator + key if parent_key else key
        if isinstance(value, Mapping):
            items.extend(
                flatten(
                    value,
                    parent_key=new_key,
                    separator=separator,
                ).items()
            )
        else:
            items.append((new_key, value))
    return dict(items)



from __future__ import annotations

import contextlib
import functools
import hashlib
import inspect
import itertools
import os
import shutil
import tempfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pyarrow as pa
import pyarrow.dataset as ds
from numpy.typing import ArrayLike

from octoflow import config
from octoflow.data.base import PyArrowWrapper
from octoflow.data.compute import Expression

try:
    import pandas as pd
    from pandas import DataFrame as DataFrameType
except ImportError:
    pd = None
    DataFrameType = None


DatasetType = PyArrowWrapper[ds.Dataset]
SourceType = Union[str, List[str], Union[Path, List[Path]], "Dataset", List["Dataset"]]

DEFAULT_BATCH_SIZE = 131_072
DEFAULT_FORMAT = "arrow"


def mapfunc(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        output = func(*args, **kwargs)
        if isinstance(output, pd.Series):
            return output
        return pd.Series(output)

    return wrapped


def get_data_path(path: Union[Path, str]) -> Path:
    if isinstance(path, str):
        path = Path(path)
    if path.is_dir() or not path.exists():
        path.mkdir(
            parents=True,
            exist_ok=True,
        )
        data_path = path / "data"
    else:
        msg = f"expected path to be directory, got '{path}'"
        raise ValueError(msg)
    return data_path


class Dataset(DatasetType):
    def __init__(
        self,
        data: Union[List[dict], Dict[str, list], DataFrameType] = None,
        path: Optional[Union[str, Path]] = None,
        format: str = DEFAULT_FORMAT,
    ):
        # private attributes
        if path is None:
            # create a temporary directory in system temp directory
            cache_dir = Path(config.resources.path).expanduser() / "cache" / "datasets"
            if not cache_dir.exists():
                cache_dir.mkdir(parents=True, exist_ok=True)
            path = Path(tempfile.mkdtemp(dir=cache_dir))
        data_path = get_data_path(path)
        self._path = path
        self._format = format
        ds.write_dataset(pa.table(data), data_path, format=format)
        dataset: ds.Dataset = ds.dataset(data_path, format=self._format)
        super().__init__(dataset)

    @classmethod
    def load_dataset(
        cls,
        path: Union[Path, str],
        format: str = DEFAULT_FORMAT,
    ) -> Dataset:
        # create instance using __new__
        inst = cls.__new__(cls)
        inst._path = path
        inst._format = format
        inst._wrapped = ds.dataset(
            get_data_path(path),
            format=inst._format,
        )
        return inst

    @property
    def path(self) -> Path:
        return self._path

    @property
    def format(self) -> str:
        return self._format

    @property
    def _wrapped_format_default_extname(self) -> str:
        return self._wrapped.format.default_extname

    def count_rows(self) -> int:
        return self._wrapped.count_rows()

    def head(
        self,
        num_rows: int = 5,
        columns: Union[str, List[str], None] = None,
        filter: Expression = None,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> DataFrameType:
        filter = filter.to_pyarrow() if filter else None
        if isinstance(columns, str):
            columns = [columns]
        table: pa.Table = self._wrapped.head(
            num_rows=num_rows,
            columns=columns,
            filter=filter,
            batch_size=batch_size,
        )
        return table.to_pandas()

    def take(
        self,
        indices: Union[int, slice, List[int], ArrayLike] = [],
        columns: Union[str, List[str], None] = None,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> DataFrameType:
        """
        Take rows from the dataset.

        Parameters
        ----------
        indices : int, slice, list of int, array-like
            Indices of rows to take.
        columns : str, list of str, None
            Names of columns to take. If None, all columns are taken.
        batch_size : int
            Number of rows to take at a time.

        Returns
        -------
        DataFrameType
            A pandas DataFrame containing the taken rows.
        """
        if isinstance(indices, int):
            indices = [indices]
        elif isinstance(indices, slice):
            step_size = indices.step
            if step_size is None:
                # default step size is 1
                step_size = 1
            elif not isinstance(step_size, int):
                msg = f"expected indices.step to be int, got '{indices.step}'"
                raise ValueError(msg)
            elif step_size < 1:
                msg = f"expected indices.step to be greater than 0, got '{step_size}'"
                raise ValueError(msg)
            start = indices.start or 0
            stop = indices.stop or self.count_rows()
            indices = np.arange(start, stop, step_size)
        if isinstance(columns, str):
            columns = [columns]
        table: pa.Table = self._wrapped.take(
            indices=indices,  # array or array-like
            columns=columns,  # list of str or None
            batch_size=batch_size,  # int
        )
        return table.to_pandas()

    def __getitem__(self, indices: Union[int, slice, List[int]]) -> Dataset:
        return self.take(indices)

    def _sync_pyarrow_dataset(
        self,
        suffix: Union[str, Path, None],
        data: Union[
            ds.Dataset,
            pa.RecordBatch,
            pa.Table,
            List[pa.RecordBatch],
            List[pa.Table],
        ],
        schema: pa.Schema = None,
    ) -> Dataset:
        out_path = self.path if suffix is None else self.path / suffix
        if not out_path.exists():
            # first write to temporary directory
            temp_path = Path(tempfile.mkdtemp(dir=self.path / ".tmp"))
            ds.write_dataset(
                data,
                temp_path,
                schema=schema,
                format=self.format,
            )
            out_data_path = get_data_path(out_path)
            with contextlib.suppress(OSError):
                os.replace(temp_path, out_data_path)
        if suffix is None:
            return self
        return type(self).load_dataset(out_path)

    def map(
        self,
        func: Any,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> Dataset:
        batch_iter = (
            pa.RecordBatch.from_pandas(
                batch.to_pandas().apply(
                    mapfunc(func),
                    axis=1,
                ),
            )
            for batch in self._wrapped.to_batches(batch_size=batch_size)
        )
        try:
            first: pa.RecordBatch = next(batch_iter)
        except StopIteration:
            return self
        batch_iter = itertools.chain([first], batch_iter)
        schema = None if first is None else first.schema
        func_hash = hashlib.sha256(inspect.getsource(func).encode()).hexdigest()
        return self._sync_pyarrow_dataset(f"mapped-{func_hash}", batch_iter, schema)

    def filter(self, expression: Expression = None) -> Dataset:
        if expression is None:
            return self
        pyarrow_expression = expression.to_pyarrow()
        dataset = self._wrapped.filter(pyarrow_expression)
        expr_hash = hashlib.sha256(str(pyarrow_expression).encode()).hexdigest()
        return self._sync_pyarrow_dataset(f"filter-{expr_hash}", dataset, dataset.schema)

    def cleanup(self):
        if not self.path.exists():
            return
        shutil.rmtree(self.path)

    def __enter__(self) -> Dataset:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        with contextlib.suppress(Exception):
            self.cleanup()
